from agent.SerialLineReceiverBase import SerialLineReceiverBase
from agent.BaseContext import BaseContext
import platform
import logging
from twisted.internet.task import LoopingCall
from twisted.internet import reactor, defer
import uptime
from inspect import getfullargspec, isfunction

logger = logging.getLogger()


class SerialLineReceiver(SerialLineReceiverBase):

    def connectionMade(self):
        self.transport.write(b'\r\n')
        self.send_command(
            command='agent_started',
            args={'version': self.context.get_agent_version(), 'system': platform.system()})

        def shutdown():
            self.connectionLost2('shutdown')
            d = defer.Deferred()
            reactor.callLater(0.3, d.callback, "1")
            return d
        reactor.addSystemEventTrigger("before", "shutdown", shutdown)

    def connectionLost(self, reason):
        reactor.stop()

    def connectionLost2(self, reason):
        self.send_command(command='agent_stopped', args={})

    def tick(self):
        logger.debug("Sending tick")
        try:
            self.send_status()
        except Exception:
            logger.exception("Twisted hide exception")

    def __init__(self, context: BaseContext):
        super(SerialLineReceiver, self).__init__()
        self.lc = LoopingCall(self.tick)
        self.lc.start(5, now=False)
        self.context = context

    def send_status(self):
        import psutil
        disk_usage = dict((disk.device.replace('/', '_'),
                           psutil.disk_usage(disk.mountpoint).percent)
                          for disk in psutil.disk_partitions())
        args = {"cpu": dict(psutil.cpu_times()._asdict()),
                "ram": dict(psutil.virtual_memory()._asdict()),
                "swap": dict(psutil.swap_memory()._asdict()),
                "uptime": {"seconds": uptime.uptime()},
                "disk": disk_usage,
                "user": {"count": len(psutil.users())}}
        self.send_response(response='status', args=args)
        logger.debug("send_status finished")

    def _check_args(self, func, args):
        if not isinstance(args, dict):
            raise TypeError("Arguments should be all keyword-arguments in a "
                            "dict for command %s instead of %s." %
                            (self._pretty_fun(func), type(args).__name__))

        # check for unexpected keyword arguments
        argspec = getfullargspec(func)
        if argspec.varkw is None:  # _operation doesn't take ** args
            unexpected_kwargs = set(args) - set(argspec.args)
            if unexpected_kwargs:
                raise TypeError(
                    "Command %s got unexpected keyword arguments: %s" % (
                        self._pretty_fun(func), ", ".join(unexpected_kwargs)))

        mandatory_args = argspec.args
        if argspec.defaults:  # remove those with default value
            mandatory_args = mandatory_args[0:-len(argspec.defaults)]
        missing_kwargs = set(mandatory_args) - set(args)
        if missing_kwargs:
            raise TypeError("Command %s missing arguments: %s" % (
                self._pretty_fun(func), ", ".join(missing_kwargs)))

    def _get_command(self, command, args):
        if not isinstance(command, str) or command.startswith('_'):
            raise AttributeError('Invalid command: %s' % command)
        try:
            func = getattr(self.context, command)
        except AttributeError as e:
            raise AttributeError('Command not found: %s (%s)' % (command, e))

        if not isfunction(func):
            raise AttributeError("Command refers to non-static method %s." % self._pretty_fun(func))

        self._check_args(func, args)
        logger.debug("_get_command finished")
        return func

    @staticmethod
    def _pretty_fun(fun):
        try:
            argspec = getfullargspec(fun)
            args = argspec.args
            if argspec.varargs:
                args.append("*" + argspec.varargs)
            if argspec.varkw:
                args.append("**" + argspec.varkw)
            return "%s(%s)" % (fun.__name__, ",".join(args))
        except Exception:
            return "<%s>" % type(fun).__name__

    def handle_command(self, command, args):
        func = self._get_command(command, args)
        retval = func(**args)
        self.send_response(
            response=func.__name__,
            args={'retval': retval, 'uuid': args.get('uuid', None)})

    def handle_response(self, response, args):
        pass
